/**************************************************************************
 * This is the source code of the HPTTopTagger (version 2)                *
 * as used in arXiv:1505.01921                                            *
 **************************************************************************
 *
 * If you use it please cite
 *
 * %\cite{Spannowsky:2015eba}
 * \bibitem{Spannowsky:2015eba} 
 * M.~Spannowsky and M.~Stoll,
 * %``Tracking New Physics at the LHC and beyond,''
 * arXiv:1505.01921 [hep-ph].
 * %%CITATION = ARXIV:1505.01921;%%
 *
 * and also the original HPTTopTagger 
 *
 * %\cite{Schaetzel:2013vka}
 * \bibitem{Schaetzel:2013vka}
 * S.~Schaetzel and M.~Spannowsky,
 * %``Tagging highly boosted top quarks,''
 * Phys.\ Rev.\ D {\bf 89}, 014007 (2014)
 * [arXiv:1308.0540 [hep-ph]].
 * %%CITATION = ARXIV:1308.0540;%%
 *
 * due to the similarities with the HEPTopTagger algorithm please consider
 * also citing
 *
 * %\cite{Plehn:2010st}
 * \bibitem{Plehn:2010st}
 * T.~Plehn, M.~Spannowsky, M.~Takeuchi and D.~Zerwas,
 * %``Stop Reconstruction with Tagged Tops,''
 * JHEP {\bf 1010}, 078 (2010)
 * [arXiv:1006.2833 [hep-ph]].
 * %%CITATION = ARXIV:1006.2833;%%
 ***/

/**************************************************************************
 * basic usage of the HPTTopTagger                                        *
 **************************************************************************
 *
 * The constructor uses the following arguments:
 *  1. the ClusterSequence of the track/e-cal-based fat jet
 *  2. the corresponding fat jet
 *  3. the ClusterSequence of the calorimeter-based fat jet
 *  4. the corresponding fat jet
 * optional:
 *  5. top quark mass
 *  6. W boson mass
 *
 * run_tagger()
 * is_tagged(): true / false
 * top_candidate(): returns the top candidate
 ***/

#ifndef __HPTTOPTAGGER_HH__
#define __HPTTOPTAGGER_HH__

using namespace std;

class HPTTopTagger {
public:

  HPTTopTagger(const fastjet::ClusterSequence & ccs,
	       const fastjet::PseudoJet & cjet,
	       const fastjet::ClusterSequence & cs,
	       const fastjet::PseudoJet & jet);

  HPTTopTagger(const fastjet::ClusterSequence & ccs,
	       const fastjet::PseudoJet & cjet,
	       const fastjet::ClusterSequence & cs,
	       const fastjet::PseudoJet & jet,
	       double mtmass, double mwmass);

  // run and get result
  void run_tagger();
  bool is_maybe_top() const {return _is_maybe_top;}
  bool is_masscut_passed() const {return _is_masscut_passed;}
  bool is_wmass_passed() const {return _is_wmass_passed;}
  bool is_ptr_cut_passed() const {return _is_ptr_cut_passed;}
  bool is_tagged() const {return _is_masscut_passed && _is_wmass_passed &&
      _is_ptr_cut_passed;}
  const PseudoJet & top_candidate() const {return _top_candidate;}
  const PseudoJet & W_candidate() const {return _W_candidate;}
  const vector<fastjet::PseudoJet> & top_subjets() const {return _top_subjets;}
  const vector<fastjet::PseudoJet> & top_hadrons() const {return _top_hadrons;}
  const vector<fastjet::PseudoJet> & top_charged_constits() const
  {return _top_charged_constits;}
  const vector<fastjet::PseudoJet> & hardparts() const {return _top_parts;}
  unsigned parts_size() const {return _parts_size;}
  double delta_top() const {return _delta_top;}
  double drmin() const {return _drmin;}
  double drmin3() const {return _drmin3;}

  // get info
  void get_setting() const;
  void get_info() const;

  // set parameters
  void set_top_range(double xmin, double xmax) {_mtmin=xmin; _mtmax=xmax;}
  void set_W_range(double xmin, double xmax) {_mwmin=xmin; _mwmax=xmax;}
  void set_mass_ratio_range(double rmin, double rmax) {_rmin=rmin; _rmax=rmax;}
  void set_mass_ratio_cut(double m23cut, double m13cutmin,double m13cutmax)
  {_m23cut=m23cut; _m13cutmin=m13cutmin; _m13cutmax=m13cutmax;}
  void set_alpha_R ( double r ) { _alpha_R = r; }
  void set_subjet_RR ( double rr ) { _subjet_RR = rr; }
  void set_recl_relative_r ( double r ) { _recl_relative_r = r; }

  bool debugg;

private:
  const fastjet::ClusterSequence * _cs;
  const fastjet::ClusterSequence * _ccs;
  const fastjet::PseudoJet _jet;
  const fastjet::PseudoJet _cjet;
  const double _mtmass, _mwmass;
  double _mtmin, _mtmax;
  double _mwmin, _mwmax;
  double _rmin, _rmax;
  double _m23cut, _m13cutmin, _m13cutmax;
  double _drmin, _drmin3;
  double _alpha_R, _subjet_RR, _recl_relative_r;
  double _pt_times_r_min, _pt_times_r_max;

  bool _is_maybe_top;
  bool _is_masscut_passed;
  bool _is_wmass_passed;
  bool _is_ptr_cut_passed;
  double _delta_top;
  unsigned _parts_size;
  fastjet::PseudoJet _top_candidate;
  fastjet::PseudoJet _W_candidate;
  vector<fastjet::PseudoJet> _top_subjets;
  vector<fastjet::PseudoJet> _top_hadrons;
  vector<fastjet::PseudoJet> _top_charged_constits;

  vector<fastjet::PseudoJet> _top_parts;
  //vector<fastjet::PseudoJet> _top_charged_parts;

  vector<fastjet::PseudoJet> rescale_tracks();
  vector<fastjet::PseudoJet> recluster_with_ghosts
  ( const fastjet::JetDefinition & jet_def_recl,
    const vector<fastjet::PseudoJet> & tracks,
    const vector<fastjet::PseudoJet> & hadrons,
    vector<fastjet::PseudoJet> & track_constituents,
    vector<fastjet::PseudoJet> & ass_constituents );
  void store_topsubjets(const vector<fastjet::PseudoJet>& top_subs);
  bool check_mass_criteria(const vector<fastjet::PseudoJet> & top_subs) const;
  bool check_Wmass(const vector<fastjet::PseudoJet> & top_subs,
		   fastjet::PseudoJet & Wcand) const;
  bool check_ptr_cut(const vector<fastjet::PseudoJet>& top_subs);
  double r_max_3jets(const fastjet::PseudoJet & jet1,
		     const fastjet::PseudoJet & jet2,
		     const fastjet::PseudoJet & jet3) const;
  double r_min_subjets(const vector<fastjet::PseudoJet> & jets) const;
  fastjet::PseudoJet Sum(const vector<fastjet::PseudoJet>& );
};

//-----------------------------------------------------------------------//

/********************
 * helper functions *
 ********************/

double HPTTopTagger::r_min_subjets ( const vector<fastjet::PseudoJet> & jets )
  const {
  double drmin(10000);
  for(unsigned ii=0;ii<jets.size();ii++)
    for(unsigned jj=ii+1; jj<jets.size();jj++)
      if(sqrt(jets[ii].squared_distance(jets[jj])) < drmin) drmin = sqrt(jets[ii].squared_distance(jets[jj]));

  return(drmin);
}

double HPTTopTagger::r_max_3jets ( const fastjet::PseudoJet & jet1,
				   const fastjet::PseudoJet & jet2,
				   const fastjet::PseudoJet & jet3 ) const {
  fastjet::PseudoJet jet12,jet13,jet23;
  jet12=jet1+jet2;
  jet13=jet1+jet3;
  jet23=jet2+jet3;

  double a=sqrt(jet1.squared_distance(jet2));
  double b=sqrt(jet2.squared_distance(jet3));
  double c=sqrt(jet3.squared_distance(jet1));
  double dR1=a,dR2=a;

  if(a<=b && a<=c){
    dR1=a;
    dR2=sqrt(jet12.squared_distance(jet3));
  }
  if(b<a && b<=c){
    dR1=b;
    dR2=sqrt(jet23.squared_distance(jet1));
  }
  if(c<a && c<b){
    dR1=c;
    dR2=sqrt(jet13.squared_distance(jet2));
  }
  return max(dR1,dR2);
}

void HPTTopTagger::store_topsubjets
( const vector<fastjet::PseudoJet>& top_subs ) {
  _top_subjets.resize(0);
  double m12=(top_subs[0]+top_subs[1]).m();
  double m13=(top_subs[0]+top_subs[2]).m();
  double m23=(top_subs[1]+top_subs[2]).m();
  double dm12=abs(m12-_mwmass);
  double dm13=abs(m13-_mwmass);
  double dm23=abs(m23-_mwmass);
  if(dm23<=dm12 && dm23<=dm13){
    _top_subjets.push_back(top_subs[0]); //supposed to be b
    _top_subjets.push_back(top_subs[1]); //W-jet 1
    _top_subjets.push_back(top_subs[2]); //W-jet 2
  }
  else if(dm13<=dm12 && dm13<dm23){
    _top_subjets.push_back(top_subs[1]); //supposed to be b
    _top_subjets.push_back(top_subs[0]); //W-jet 1
    _top_subjets.push_back(top_subs[2]); //W-jet 2
  }
  else if(dm12<dm23 && dm12<dm13){
    _top_subjets.push_back(top_subs[2]); //supposed to be b
    _top_subjets.push_back(top_subs[0]); //W-jet 1
    _top_subjets.push_back(top_subs[1]); //W-jet 2
  }
  return;
}

bool HPTTopTagger::check_mass_criteria
( const vector<fastjet::PseudoJet> & top_subs) const {
  bool is_passed=false;
  double m12=(top_subs[0]+top_subs[1]).m();
  double m13=(top_subs[0]+top_subs[2]).m();
  double m23=(top_subs[1]+top_subs[2]).m();
  double m123=(top_subs[0]+top_subs[1]+top_subs[2]).m();
  if( (atan(m13/m12)>_m13cutmin && _m13cutmax > atan(m13/m12)
       && (m23/m123>_rmin && _rmax>m23/m123))
      ||
      (((m23/m123)*(m23/m123) < 1-_rmin*_rmin*(1+(m13/m12)*(m13/m12))) &&
       ((m23/m123)*(m23/m123) > 1-_rmax*_rmax*(1+(m13/m12)*(m13/m12))) &&
       (m23/m123 > _m23cut))
      ||
      (((m23/m123)*(m23/m123) < 1-_rmin*_rmin*(1+(m12/m13)*(m12/m13))) &&
       ((m23/m123)*(m23/m123) > 1-_rmax*_rmax*(1+(m12/m13)*(m12/m13))) &&
       (m23/m123 > _m23cut))
      ) {
    is_passed=true;
  }
  return is_passed;
}

bool HPTTopTagger::check_Wmass ( const vector<fastjet::PseudoJet> & top_subs,
				 fastjet::PseudoJet & Wcand) const {
  bool is_passed=false;
  double tmpdeltaW(100000.0);

  for(unsigned rr=0; rr<top_subs.size(); rr++) {
    for(unsigned ll=rr+1; ll<top_subs.size(); ll++) {
      double deltaW = abs((top_subs[rr]+top_subs[ll]).m() - _mwmass);
      if(deltaW < tmpdeltaW) {
	tmpdeltaW = deltaW;
	Wcand = top_subs[rr]+top_subs[ll];
      }
    }
  }

  if((Wcand.m()<_mwmax) && (Wcand.m()>_mwmin)) {
    is_passed=true;
  }
  return is_passed;
}

bool HPTTopTagger::check_ptr_cut(const vector<fastjet::PseudoJet>& top_subs) {
  double pt_times_r = Sum(top_subs).pt() *
    r_max_3jets(top_subs[0],top_subs[1],top_subs[2]);
  return pt_times_r >= _pt_times_r_min && pt_times_r <= _pt_times_r_max;
}

vector<fastjet::PseudoJet> HPTTopTagger::rescale_tracks() {
  // add ghosts to hadrons
  vector<fastjet::PseudoJet> tracks = _ccs->constituents(_cjet),
    ghosts;
  for ( unsigned i=0; i<tracks.size(); ++i ) {
    fastjet::PseudoJet g;
    g.reset_PtYPhiM ( 0.00001, tracks[i].eta(), tracks[i].phi(),
		      0.00000001 );
    g.set_user_index(i);
    ghosts.push_back(g);
  }
  vector<fastjet::PseudoJet> hadrons_ghosts = _cs->constituents(_jet);
  hadrons_ghosts.insert ( hadrons_ghosts.end(), ghosts.begin(), ghosts.end() );

  // recluster
  fastjet::JetDefinition sub_jet_def(fastjet::antikt_algorithm, _alpha_R);
  fastjet::ClusterSequence sub_clust_seq(hadrons_ghosts, sub_jet_def);
  vector<fastjet::PseudoJet> subjets = sub_clust_seq.inclusive_jets();

  // rescale
  vector<fastjet::PseudoJet> rescaled_tracks;
  for ( unsigned i=0; i<subjets.size(); ++i ) {
    vector<fastjet::PseudoJet> subjet_tracks,
      subjet_constits = subjets[i].constituents();
    for ( unsigned j=0; j<subjet_constits.size(); ++j ) {
      if ( fabs(subjet_constits[j].pt() - 0.00001) < 1e-8 )
	subjet_tracks.push_back ( tracks[ subjet_constits[j].user_index() ] );
    }
    double alpha = subjets[i].e() / Sum(subjet_tracks).e();
    if ( alpha < 1. ) alpha = 1.;
    for ( unsigned j=0; j<subjet_tracks.size(); ++j )
      rescaled_tracks.push_back ( alpha * subjet_tracks[j] );
  }

  return rescaled_tracks;
}

vector<fastjet::PseudoJet> HPTTopTagger::recluster_with_ghosts
( const fastjet::JetDefinition & jet_def_recl,
  const vector<fastjet::PseudoJet> & tracks,
  const vector<fastjet::PseudoJet> & hadrons,
  vector<fastjet::PseudoJet> & track_constituents,
  vector<fastjet::PseudoJet> & ass_constituents ) {

  // add ghosts to tracks
  vector<fastjet::PseudoJet> ghosts;
  for ( unsigned i=0; i<hadrons.size(); ++i ) {
    fastjet::PseudoJet g;
    g.reset_PtYPhiM ( 0.00001, hadrons[i].eta(), hadrons[i].phi(),
		      0.00000001 );
    g.set_user_index(i);
    ghosts.push_back(g);
  }
  vector<fastjet::PseudoJet> tracks_ghosts = tracks;
  tracks_ghosts.insert ( tracks_ghosts.end(), ghosts.begin(), ghosts.end() );

  // recluster
  fastjet::ClusterSequence clust_seq_recl(tracks_ghosts, jet_def_recl);
  vector<fastjet::PseudoJet> jets = clust_seq_recl.inclusive_jets();

  // separate tracks and ghosts
  track_constituents.clear(); ass_constituents.clear();
  for ( unsigned i=0; i<jets.size(); ++i ) {
    vector<fastjet::PseudoJet> constits = jets[i].constituents();
    for ( unsigned j=0; j<constits.size(); ++j ) {
      if ( fabs(constits[j].pt() - 0.00001) < 1e-8 ) {
	ass_constituents.push_back ( hadrons[ constits[j].user_index() ] );
	jets[i] -= constits[j]; // remove ghost from jet
      }
      else {
	track_constituents.push_back ( constits[j] );
      }
    }
  }

  return jets;
}

fastjet::PseudoJet HPTTopTagger::Sum
( const vector<fastjet::PseudoJet> & vec_pjet ) {

  fastjet::PseudoJet sum;
  sum.reset(0.,0.,0.,0.);
  for(unsigned i=0;i<vec_pjet.size();i++){
    sum += vec_pjet.at(i);
  }
  return sum;
}


/****************
 * HPTTopTagger *
 ****************/

HPTTopTagger::HPTTopTagger ( const fastjet::ClusterSequence & ccs,
			     const fastjet::PseudoJet & cjet,
			     const fastjet::ClusterSequence & cs,
			     const fastjet::PseudoJet & jet) :
  debugg(false),
  _cs(&cs), _ccs(&ccs), _jet(jet), _cjet(cjet), _mtmass(173.2), _mwmass(80.4),
  _mtmin(173.2 - 25.),_mtmax(173.2 + 25.), _mwmin(80.4-15.),_mwmax(80.4+15.),
  _rmin(0.85*80.4/173.2),_rmax(1.15*80.4/173.2),
  _m23cut(0.35),_m13cutmin(0.2),_m13cutmax(1.3),
  _alpha_R(0.2), _subjet_RR(100.), _recl_relative_r(0.8),
  _pt_times_r_min(250.), _pt_times_r_max(750.)
{}

HPTTopTagger::HPTTopTagger ( const fastjet::ClusterSequence & ccs,
			     const fastjet::PseudoJet & cjet,
			     const fastjet::ClusterSequence & cs,
			     const fastjet::PseudoJet & jet,
			     double mtmass, double mwmass) :
  debugg(false),
  _cs(&cs), _ccs(&ccs), _jet(jet), _cjet(cjet),
  _mtmass(mtmass), _mwmass(mwmass),
  _mtmin(mtmass - 25.),_mtmax(mtmass + 25.),
  _mwmin(mwmass-15.),_mwmax(mwmass+15.),
  _rmin(0.85*mwmass/mtmass),_rmax(1.15*mwmass/mtmass),
  _m23cut(0.35),_m13cutmin(0.2),_m13cutmax(1.3),
  _alpha_R(0.2), _subjet_RR(100.), _recl_relative_r(0.8),
  _pt_times_r_min(250.), _pt_times_r_max(750.)
{}
    
void HPTTopTagger::run_tagger()
{
  _delta_top=1000000000000.0;
  _top_candidate.reset(0.,0.,0.,0.);
  _W_candidate.reset(0.,0.,0.,0.);
  _parts_size=0;
  _is_maybe_top=false;
  _is_masscut_passed=false;
  _is_wmass_passed=false;
  _top_subjets.clear();
  _top_hadrons.clear();
  _top_charged_constits.clear();
  _top_parts.clear();
  //_top_charged_parts.clear();

  if(debugg)
    {
      std::cout << "mtmass in top_tagger: " << _mtmass << std::endl;
      std::cout << "mwmass in top_tagger: " << _mwmass << std::endl;
      std::cout << "jet input HPTTopTagger: " << std::endl;
      //printjet(_jet); //TODO
    }

  vector<fastjet::PseudoJet> rescaled_tracks = rescale_tracks();
  assert ( rescaled_tracks.size() == (_ccs->constituents(_cjet)).size() );

  // make subjets
  fastjet::JetDefinition jet_def_subs ( fastjet::antikt_algorithm,
					_subjet_RR/Sum(rescaled_tracks).pt() );
  fastjet::ClusterSequence clust_seq_subs ( rescaled_tracks, jet_def_subs );
  _top_parts = clust_seq_subs.inclusive_jets();
  if(_top_parts.size() < 3) return;

  _top_parts = sorted_by_pt(_top_parts);

  // store hard substructure of the top candidate
  _parts_size=_top_parts.size();
  _drmin =r_min_subjets(_top_parts);
  _drmin3 = _top_parts.size() < 3 ? 0. :
    r_min_subjets(vector<fastjet::PseudoJet>(_top_parts.begin(),
					     _top_parts.begin()+3));

  // recluster with radius close to _drmin3,
  // including ghost hadrons
  if ( _top_parts.size() < 3 ) return;
  vector<fastjet::PseudoJet> top_charged_constits, top_hadrons;
  fastjet::JetDefinition jet_def_recl ( fastjet::cambridge_algorithm,
					_recl_relative_r * _drmin3 );
  vector<fastjet::PseudoJet> top_subs =
    sorted_by_pt ( recluster_with_ghosts( jet_def_recl,
					  rescaled_tracks,
					  _cs->constituents(_jet),
					  top_charged_constits,
					  top_hadrons ) );
  if ( top_subs.size() < 3 ) return;
  top_subs.resize(3);

  if(debugg) {
    std::cout << "number of top_charged_constits: "
	      << top_charged_constits.size() << std::endl;
  }

  // top candidate
  fastjet::PseudoJet topcandidate = Sum(top_subs);
  if(debugg)
    std::cout << "HPTTopTagger: topcandidate mass:" << topcandidate.m()
	      << std::endl;
  if( topcandidate.m() < _mtmin || _mtmax < topcandidate.m() ) return;

  // transfer infos of the positively identified top to the outer world
  double deltatop = abs(topcandidate.m() - _mtmass);
  _delta_top = deltatop;
  _is_maybe_top = true;
  _top_candidate = topcandidate;
  store_topsubjets(top_subs);
  _top_hadrons=top_hadrons;
  _top_charged_constits = top_charged_constits;
  // check all cuts
  _is_masscut_passed=check_mass_criteria(top_subs);
  _is_wmass_passed=check_Wmass(top_subs, _W_candidate);
  _is_ptr_cut_passed=check_ptr_cut(top_subs);

  return;
}

void HPTTopTagger::get_info() const
{
  std::cout << "maybe_top: " <<  _is_maybe_top << std::endl;
  std::cout << "masscut_passed: " <<  _is_masscut_passed << std::endl;
  std::cout << "wmass_passed: " <<  _is_wmass_passed << std::endl;
  std::cout << "ptr_cut_passed: " <<  _is_ptr_cut_passed << std::endl;
  std::cout << "top candidate mass:" <<  _top_candidate.m() << std::endl;
  std::cout << "top candidate (pt, eta, phi): ("
	    <<  _top_candidate.perp() << ","
	    <<  _top_candidate.eta() << ","
	    <<  _top_candidate.phi_std() << ")" << std::endl;
  std::cout << "hadrons size: " <<  _top_hadrons.size() << std::endl;
  std::cout << "charged constituents size: " << _top_charged_constits.size()
	    << std::endl;
  std::cout << "parts size: " <<  _parts_size << std::endl;
  std::cout << "delta_top: " <<  _delta_top << std::endl;
  return;
}

void HPTTopTagger::get_setting() const
{
  std::cout << "top mass: " <<  _mtmass << std::endl;
  std::cout << "W mass: " <<  _mwmass << std::endl;
  std::cout << "top mass range: [" << _mtmin << ", " << _mtmax << "]"
	    << std::endl;
  std::cout << "W mass range: [" << _mwmin << ", " << _mwmax << "]"
	    << std::endl;
  std::cout << "W mass ratio range: [" << _rmin << ", " << _rmax << "] (["
	    <<_rmin*_mtmass/_mwmass<< "%, "<< _rmax*_mtmass/_mwmass << "%])"
	    << std::endl;
  std::cout << "mass ratio cut: (m23cut, m13min, m13max)=("
	    << _m23cut << ", " << _m13cutmin << ", " << _m13cutmax << ")"
	    << std::endl;
  std::cout << "track calibration alpha_R: " << _alpha_R << std::endl;
  std::cout << "subjet_RR: " << _subjet_RR << std::endl;
  std::cout << "recl_relative_r: " << _recl_relative_r << std::endl;
  return;
}


#endif // __HPTTOPTAGGER_HH__
